#ident "%W% %E% Copyright (c) Gert Doering"
/* flow.c - handle flow control stuff */

#define FLOW_HARD	0x01		/* rts/cts */
#define FLOW_XON_IN	0x02		/* incoming data, send xon/xoff */
#define FLOW_XON_OUT	0x04		/* send data, honor xon/xoff */
 
#ifdef POSIX_TERMIOS
#include <termios.h>
#else
#include <termio.h>
#endif

#include "mgetty.h"

/* hardware handshake flags - use what is available on local system
 */

#ifdef CRTSCTS
# define HARDWARE_HANDSHAKE CRTSCTS		/* linux, SunOS */
#else
# ifdef CRTSFL
#  define HARDWARE_HANDSHAKE CRTSFL		/* SCO 3.2v4.2 */
# else
#  ifdef RTSFLOW
#   define HARDWARE_HANDSHAKE RTSFLOW | CTSFLOW	/* SCO 3.2v2 */
#  else
#   ifdef CTSCD
#    define HARDWARE_HANDSHAKE CTSCD		/* AT&T 3b1? */
#   else
#    define HARDWARE_HANDSHAKE 0		/* nothing there... */
#   endif
#  endif
# endif
#endif

/* Dial-Out parallel to a Dial-in on SCO 3.2v4.0 does only work if
 * *only* CTSFLOW is set (Uwe S. Obst)
 */
#ifdef BROKEN_SCO_324
#undef HARDWARE_HANDSHAKE
#define HARDWARE_HANDSHAKE CTSFLOW
#endif

int set_flow_control( int fd, int type )
{
#ifdef POSIX_TERMIOS
struct termios t;
#else
struct termio  t;
#endif

#ifdef POSIX_TERMIOS
    if ( tcgetattr( fd, &t ) < 0 )
    {
	lprintf( L_ERROR, "set_flow_control: tcgetattr" );
	return -1;
    }
#else
    if ( ioctl( fd, TCGETA, &t ) < 0 )
    {
	lprintf( L_ERROR, "set_flow_control: TCGETA" );
	return -1;
    }
#endif
    t.c_cflag &= ~HARDWARE_HANDSHAKE;
    t.c_iflag &= ~( IXON | IXOFF | IXANY );

    if ( type & FLOW_HARD )
			t.c_cflag |= HARDWARE_HANDSHAKE;
    if ( type & FLOW_XON_IN )
			t.c_iflag |= IXOFF;
    if ( type & FLOW_XON_OUT )
			t.c_iflag |= IXON | IXANY;

#ifdef POSIX_TERMIOS
    if ( tcsetattr( fd, TCSANOW, &t ) < 0 )
    {
	lprintf( L_ERROR, "set_flow_control: tcsetattr" );
	return -1;
    }
#else
    if ( ioctl( fd, TCSETA, &t ) < 0 )
    {
	lprintf( L_ERROR, "set_flow_control: TCSETA" );
	return -1;
    }
#endif
    return 0;
}
