#ident "@(#)g3cat.c	1.3 93/10/03 (c) Gert Doering";

/* g3cat.c - concatenate multiple G3-Documents
 *
 * Syntax: g3cat [options] [file(s) (or "-" for stdin)]
 *
 * Valid options: -l (separate g3 files with a black line)
 *                -d (output digifax header)
 *                -a (byte-align EOLs)
 */

/* #define DEBUG 1 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

int byte_align = 0;

static int pbit = 0x80;
static unsigned char buf[2000];
static int buflen = 0;

#ifdef __GNUC__
inline
#endif
void putbit0( int number )		/* write "number" 0-bits */
{
    while ( number > 0 )
    {
	pbit >>= 1;
	if ( pbit == 0 )
	{
	    buflen++; buf[buflen] = 0; pbit = 0x80;
	}
	number--;
    }
    if ( buflen > sizeof( buf ) - 10 )
    {
	write( 1, buf, buflen );
	buf[ 0 ] = buf[ buflen ];
	buflen = 0;
    }
}

#ifdef __GNUC__
inline
#endif
void putbit1( void )			/* write one 1-bit */
{
    buf[ buflen ] |= pbit;
    pbit >>= 1;
    if ( pbit == 0 )
    {
	buflen++; buf[buflen] = 0 ; pbit = 0x80;
    }
}

#ifdef __GNUC__
inline
#endif
void puteol( void )			/* write byte-aligned EOL */
{
    if ( byte_align ) while( pbit != 0x08 ) putbit0( 1 );
    putbit0( 11 );
    putbit1();
}

void putbitstring( unsigned long bits, int length )
{
unsigned long b = ( 1 << (length-1) );

    while ( b != 0 )
    {
	if ( ( bits & b) == 0 ) putbit0( 1 );
			   else putbit1();
	b >>= 1;
    }
}

int putblackline = 0;          /* do not output black line */	

int main( int argc, char ** argv )
{
int i;
int fd;
int c;
int eols;
int ceol;
int raw0;
int bit;

char rbuf[2048];		/* read buffer */
int  r;				/* read buffer pointer */
int  rs;			/* actually read data */

    puteol();

    for ( i=1; i<argc; i++ )
    {
        if ( strcmp( argv[i], "-l" ) == 0)
        {
	    putblackline = 1; continue ;
        }
	if ( strcmp( argv[i], "-" ) == 0 )
		fd = 0;
	else
		fd = open( argv[i], O_RDONLY );

	if ( fd == -1 ) { perror( argv[i] ); continue; }

	eols = 0;
	ceol = 1;	/* there has been an EOL already */

	rs = read( fd, rbuf, sizeof(rbuf) );
	if ( rs < 0 ) { perror( "read" ); close( rs ); continue; }

			/* skip GhostScript header */
	r = ( rs >= 64 && strcmp( rbuf+1, "PC Research, Inc" ) == 0 ) ? 64 : 0;

	c = 0; bit = 0;

	raw0 = 0;

	while( rs != 0 && ceol < 5 )
	{
	    if ( bit == 0 )
	    {
		c = rbuf[ r++ ];

		if ( r >= rs )
		{
		    rs = read( fd, rbuf, sizeof( rbuf ) );
		    if ( rs < 0 ) { perror( "read2"); break; }
		    r = 0;
		}
		bit = 0x80;
	    }

	    if ( ( c & bit ) == 0 )			/* 0 bit */
	    {
		raw0++;
	    }
	    else					/* 1 bit */
	    {
		if ( raw0 >= 11 ) { 			/* found EOL */
		    if ( ceol == 0 ) puteol();
		    eols++;
		    ceol++;
		}
		else		/* not eol -> write 0 bits and 1 bit */
		{
		    ceol=0;
		    if ( raw0 > 0 ) putbit0( raw0 );
		    putbit1();
		}
		raw0=0;
	    }
	    bit >>= 1;
	}

#ifdef DEBUG
	fprintf( stderr, "%s: number of EOLs: %d\n", argv[i], eols );
#endif

	if ( fd != 0 ) close( fd );

	/* separate multiple files with a line */
	if ( i != argc -1 )
	{
	    putbitstring( 0x00009b35, 17 ); puteol();	/* white line */
            if ( putblackline )                         /* black line */
		{ putbitstring( 0x1a819437, 31 ); puteol(); }
	    putbitstring( 0x00009b35, 17 ); puteol();	/* white line */
	}

    }	/* end for (all arguments) */

    /* output final RTC */
    for ( i=0; i<6; i++ ) puteol();

    /* flush buffer */
    write( 1, buf, buflen+1 );

    return 0;
}
